/* DeScumm - Scumm Script Disassembler
 * Copyright (C) 2001  Ludvig Strigeus
 * Copyright (C) 2002-2004  The ScummVM Team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/tools/descumm-tool.cpp,v 1.9 2004/10/01 18:16:41 fingolfin Exp $
 *
 */

#include "descumm.h"

int (*pt2Print)(const char *format, ...);

int skipVerbHeader_V12(byte *p)
{
	byte code;
	int offset = 15;
	int minOffset = 255;
	p += offset;

	pt2Print("Events:\n");

	while ((code = *p++) != 0) {
		offset = *p++;
		pt2Print("  %2X - %.4X\n", code, offset);
		if (minOffset > offset)
			minOffset = offset;
	}
	return minOffset;
}

int skipVerbHeader_V34(byte *p)
{
	byte code;
	int offset = GF_UNBLOCKED ? 17 : 19;
	int minOffset = 255;
	p += offset;
	
	pt2Print("Events:\n");

	while ((code = *p++) != 0) {
		offset = TO_LE_16(*(uint16 *)p);
		p += 2;
		pt2Print("  %2X - %.4X\n", code, offset);
		if (minOffset > offset)
			minOffset = offset;
	}
	return minOffset;
}

int skipVerbHeader_V567(byte *p)
{
	byte code;
	int offset = 8;
	int minOffset = 255;
	p += offset;

	pt2Print("Events:\n");

	while ((code = *p++) != 0) {
		offset = TO_LE_16(*(uint16 *)p);
		p += 2;
		pt2Print("  %2X - %.4X\n", code, offset);
		if (minOffset > offset)
			minOffset = offset;
	}
	return minOffset;
}

int skipVerbHeader_V8(byte *p)
{
	uint32 *ptr;
	uint32 code;
	int offset;
	int minOffset = 255;
	
	ptr = (uint32 *)p;
	while ((code = TO_LE_32(*ptr++)) != 0) {
		offset = TO_LE_32(*ptr++);
		pt2Print("  %2d - %.4X\n", code, offset);
		if (minOffset > offset)
			minOffset = offset;
	}
	return minOffset;
}


/*
This function will parse the given memory buffer.
Return values:
0 = success
1 = scriptVersion out of bound
2 = ScriptType unknown
*/
int descummScript(byte *mem, int len, int scriptVersion, int (*pt2OutFunc)(const char *format, ...))
{
	byte *memorg;
	char *buf;
	int len;
	
	memorg = mem;
    pt2Print = pt2OutFunc;
   	size_of_code = len;

	if (scriptVersion <= 0 || scriptVersion >= 8)
		return 1;

	switch (scriptVersion) {
		case 1:
		case 2:
			GF_UNBLOCKED = true;
			g_jump_opcode = 0x18;
			break;

		case 3:
		case 4:
		case 5:
			g_jump_opcode = 0x18;
			break;

		case 6:
		case 7:
			g_jump_opcode = 0x73;
			break;

		case 8:
			g_jump_opcode = 0x66;
			break;
	}

	buf = (char *)malloc(8192);

	offs_of_line = 0;

	if (GF_UNBLOCKED) {
		if (size_of_code < 4) {
			pt2Print("File too small to be a script\n");
			return 2;
		}
		// Hack to detect verb script: first 4 bytes should be file length
		if (TO_LE_32(*((uint32 *)mem)) == size_of_code) {
			if (scriptVersion <= 2)
				offs_of_line = skipVerbHeader_V12(mem);
			else
				offs_of_line = skipVerbHeader_V34(mem );
		} else {
			mem += 4;
		}
	} else if (scriptVersion >= 5) {
		if (size_of_code < (scriptVersion == 5 ? 8 : 10)) {
			pt2Print("File too small to be a script\n");
			return 2;
		}
	
		switch (TO_BE_32(*((uint32 *)mem))) {
		case 'LSC2':
			pt2Print("Script# %d\n", TO_LE_32(*((int32 *)(mem+8))));
			mem += 12;
			break;											/* Local script */
		case 'LSCR':
			if (scriptVersion == 8) {
				pt2Print("Script# %d\n", TO_LE_32(*((int32 *)(mem+8))));
				mem += 12;
			} else if (scriptVersion == 7) {
				pt2Print("Script# %d\n", TO_LE_16(*((int16 *)(mem+8))));
				mem += 10;
			} else {
				pt2Print("Script# %d\n", (byte)mem[8]);
				mem += 9;
			}
			break;											/* Local script */
		case 'SCRP':
			mem += 8;
			break;											/* Script */
		case 'ENCD':
			mem += 8;
			break;											/* Entry code */
		case 'EXCD':
			mem += 8;
			break;											/* Exit code */
		case 'VERB':
			if (scriptVersion == 8) {
				mem += 8;
				offs_of_line = skipVerbHeader_V8(mem);
			} else
				offs_of_line = skipVerbHeader_V567(mem);
			break;											/* Verb */
		default:
			pt2Print("Unknown script type!\n");
			return 3;
		}
	} else {
		if (size_of_code < 6) {
			pt2Print("File too small to be a script\n");
			return 2;
		}
		switch (TO_LE_16(*((uint16 *)mem + 2))) {
		case MKID('LS'):
			pt2Print("Script# %d\n", (byte)mem[6]);
			mem += 7;
			break;			/* Local script */
		case MKID('SC'):
			mem += 6;
			break;			/* Script */
		case MKID('EN'):
			mem += 6;
			break;			/* Entry code */
		case MKID('EX'):
			mem += 6;
			break;			/* Exit code */
		case MKID('OC'):
			offs_of_line = skipVerbHeader_V34(mem);
			break;			/* Verb */
		default:
			pt2Print("Unknown script type!\n");
			return 3;
		}
	}

	org_pos = mem;
	cur_pos = org_pos + offs_of_line;
	len -= mem - memorg;

	while (cur_pos < mem + len) {
		byte opcode = *cur_pos;
		int j = num_block_stack;
		buf[0] = 0;
		switch (scriptVersion) {
		case 1:
		case 2:
			next_line_V12(buf);
			break;
		case 3:
		case 4:
		case 5:
			next_line_V345(buf);
			break;
		case 6:
			if (heVersion)
				next_line_HE_V72(buf);
			else
				next_line_V67(buf);
			break;
		case 7:
			next_line_V67(buf);
			break;
		case 8:
			next_line_V8(buf);
			break;
		}
		if (buf[0]) {
			writePendingElse();
			if (haveElse) {
				haveElse = false;
				j--;
			}
			outputLine(buf, offs_of_line, opcode, j);
			offs_of_line = get_curoffs();
		}
		while (indentBlock(get_curoffs())) {
			outputLine("}", offs_of_line, -1, -1);
			offs_of_line = get_curoffs();
		}
		fflush(stdout);
	}

	pt2Print("END\n");
	
/*
	if (scriptVersion >= 6 && num_stack != 0) {
		pt2Print("Stack count: %d\n", num_stack);
		if (num_stack > 0) {
			pt2Print("Stack contents:\n");
			while (num_stack) {
				buf[0] = 0;
				se_astext(pop(), buf);
				pt2Print("%s\n", buf);
			}
		}
	}
*/
	free(memorg);

	return 0;
}
